/* qag.h, qag.cc                                                 */
/* version 0.9                                                   */
/* Stefan Bauberger (bauberger@physik.uni-wuerzburg.de           */
/* last revision 4.12.95                                         */
/* ------------------------------------------------------------- */
/* This is an adaption of the integration-routine QAG from the   */
/* QUADPACK-package for C++ and for use with complex functions   */
/* or other numerical types.                                     */
/* See:                                                          */
/*   R. Piessens, E. deDoncker-Kapenga, C. Uberhuber, D. Kahaner */
/*   doubledoublepack: a Subroutine Package for Automatic Integration    */
/*   Springer Verlag, 1983.                                      */
/* ------------------------------------------------------------- */
/* Purpose: adaptive gauss-kronrod-integration                   */
/*          of real or complex functions of a real variable      */
/* ------------------------------------------------------------- */ 
/* The code has been translated with f2c and then slightly       */
/* streamlined. It runs indepently of any f2c-header-file or     */
/* library.                                                      */
/* Note:                                                         */
/*   Some effort could still be put into making the code more    */
/*   c++-like. Any user is invited to do this.                   */
/* ------------------------------------------------------------- */
/* The calling program has to provide a function:                */
/*   RESULT_TYPE foo(Double x);                                    */
/* First initialize a integration class with:                    */
/* - QAGCLASS v;                                                 */
/*         in this case the integration will be evaluated with   */
/*         standard subdivisions setting.                        */
/* - QAGCLASS v(int error_type,Double error);                      */
/*         with error_type=FIX_RELATIVE_ERROR or                 */
/*         with error_type=FIX_ABSOLUTE_ERROR                    */
/* - QAGCLASS v(Double abs_error,Double rel_error,int max_subdiv);   */
/*         the routine will then stop when one of the conditions */
/*         is fulfilled                                          */
/* Calling sequence with a and b as integration limits:          */
/* RESULT_TYPE v.integrate(RESULT_TYPE (*foo) (Double),            */
/*                         Double a,Double b);                       */
/*           a and b are the integration limits                  */
/* ------------------------------------------------------------- */

#include "qag.h"

#define Outstream cerr

#define TRUE  1
#define FALSE 0

#ifdef QUADPRECISION  // estimated:
static const doubledouble EPMACH("4.5e-32");
static const doubledouble EPMACH50("2.2e-30");
static const doubledouble ERR_EXP("1.5");
#else
#define EPMACH 2.22045e-16
#define EPMACH50 1.11e-14
#define ERR_EXP 1.5
#endif

#define UFLOW 2.22508e-308
#define OFLOW 1.79769e+308  /* Values for IEEE-arithmetic */


static Double max(Double x,Double y)
{
  if (x>y) return(x); else return(y);
}

static Double min(Double x,Double y)
{
  if (x<y) return(x); else return(y);
}

static Double xgk_15[8] =
       { SDouble(".991455371120812639206854697526"),
         SDouble(".949107912342758524526189684048"),
	 SDouble(".864864423359769072789712788641"),
         SDouble(".741531185599394439863864773281"),
	 SDouble(".586087235467691130294144838259"),
         SDouble(".405845151377397166906606412077"),
	 SDouble(".207784955007898467600689403773"),
         SDouble("0.") };

static Double wgk_15[8] =
       { SDouble(".022935322010529224963732008059"),
         SDouble(".0630920926299785532907006631892"),
	 SDouble(".104790010322250183839876322542"),
         SDouble(".14065325971552591874518959051"),
	 SDouble(".169004726639267902826583426599"),
         SDouble(".190350578064785409913256402421"),
	 SDouble(".204432940075298892414161999235"),
         SDouble(".209482141084727828012999174892") };

static Double wg_15[4] =
       { SDouble(".129484966168869693270611432679"),
         SDouble(".279705391489276667901467771424"),
	 SDouble(".381830050505118944950369775489"),
         SDouble(".417959183673469387755102040816") };

static Double xgk_21[11] =
       { SDouble(".995657163025808080735527280689003"),
         SDouble(".973906528517171720077964012084452"),
	 SDouble(".930157491355708226001207180059508"),
         SDouble(".865063366688984510732096688423493"),
	 SDouble(".780817726586416897063717578345042"),
         SDouble(".679409568299024406234327365114874"),
	 SDouble(".562757134668604683339000099272694"),
         SDouble(".433395394129247190799265943165784"),
	 SDouble(".294392862701460198131126603103866"),
         SDouble(".148874338981631210884826001129720"),
	 SDouble("0.") };

static Double wgk_21[11] =
       { SDouble(".011694638867371874278064396062192"),
         SDouble(".032558162307964727478818972459390"),
	 SDouble(".054755896574351996031381300244580"),
         SDouble(".075039674810919952767043140916190"),
	 SDouble(".093125454583697605535065465083366"),
         SDouble(".109387158802297641899210590325805"),
	 SDouble(".123491976262065851077958109831074"),
         SDouble(".134709217311473325928054001771707"),
	 SDouble(".142775938577060080797094273138717"),
         SDouble(".147739104901338491374841515972068"),
	 SDouble(".149445554002916905664936468389821") };

static Double wg_21[5] =
       { SDouble(".066671344308688137593568809893332"),
         SDouble(".149451349150580593145776339657697"),
	 SDouble(".219086362515982043995534934228163"),
         SDouble(".269266719309996355091226921569469"),
	 SDouble(".295524224714752870173892994651338") };

static Double xgk_31[16] =
       { SDouble(".998002298693397060285172840152"),
         SDouble(".987992518020485428489565718587"),
	 SDouble(".967739075679139134257347978784"),
         SDouble(".93727339240070590430775894771"),
	 SDouble(".897264532344081900882509656455"),
         SDouble(".848206583410427216200648320774"),
	 SDouble(".790418501442465932967649294818"),
         SDouble(".724417731360170047416186054614"),
	 SDouble(".650996741297416970533735895313"),
         SDouble(".570972172608538847537226737254"),
	 SDouble(".485081863640239680693655740232"),
         SDouble(".394151347077563369897207370981"),
	 SDouble(".299180007153168812166780024266"),
         SDouble(".201194093997434522300628303395"),
	 SDouble(".101142066918717499027074231447"),
         SDouble("0.") };

static Double wgk_31[16] =
       { SDouble(".00537747987292334898779205143013"),
         SDouble(".0150079473293161225383747630758"),
	 SDouble(".0254608473267153201868740010197"),
         SDouble(".0353463607913758462220379484784"),
	 SDouble(".0445897513247648766082272993733"),
         SDouble(".0534815246909280872653431472394"),
	 SDouble(".0620095678006706402851392309608"),
         SDouble(".0698541213187282587095200770992"),
	 SDouble(".0768496807577203788944327774827"),
         SDouble(".0830805028231330210382892472861"),
	 SDouble(".0885644430562117706472754436938"),
         SDouble(".0931265981708253212254868727474"),
	 SDouble(".0966427269836236785051799076276"),
         SDouble(".0991735987217919593323931734846"),
	 SDouble(".100769845523875595044946662618"),
         SDouble(".101330007014791549017374792768") };

static Double wg_31[8] =
       { SDouble(".0307532419961172683546283935772"),
         SDouble(".0703660474881081247092674164507"),
	 SDouble(".107159220467171935011869546686"),
         SDouble(".139570677926154314447804794511"),
	 SDouble(".166269205816993933553200860481"),
         SDouble(".186161000015562211026800561866"),
	 SDouble(".198431485327111576456118326444"),
         SDouble(".202578241925561272880620199968") };

static Double xgk_41[21] =
       { SDouble(".998859031588277663838315576546"),
         SDouble(".993128599185094924786122388471"),
	 SDouble(".98150787745025025919334299472"),
         SDouble(".963971927277913791267666131197"),
	 SDouble(".940822633831754753519982722212"),
         SDouble(".912234428251325905867752441203"),
	 SDouble(".878276811252281976077442995113"),
         SDouble(".839116971822218823394529061702"),
	 SDouble(".795041428837551198350638833273"),
         SDouble(".746331906460150792614305070356"),
	 SDouble(".693237656334751384805490711846"),
         SDouble(".636053680726515025452836696226"),
	 SDouble(".575140446819710315342946036586"),
         SDouble(".510867001950827098004364050955"),
	 SDouble(".443593175238725103199992213493"),
         SDouble(".373706088715419560672548177025"),
	 SDouble(".301627868114913004320555356859"),
         SDouble(".227785851141645078080496195369"),
	 SDouble(".152605465240922675505220241023"),
         SDouble(".0765265211334973337546404093988"),
	 SDouble("0.") };

static Double wgk_41[21] =
       { SDouble(".00307358371852053150121829324603"),
         SDouble(".0086002698556429421986617879501"),
	 SDouble(".0146261692569712529837879603089"),
         SDouble(".0203883734612665235980102314328"),
	 SDouble(".0258821336049511588345050670962"),
         SDouble(".0312873067770327989585431193238"),
	 SDouble(".0366001697582007980305572407072"),
         SDouble(".0416688733279736862637883059369"),
	 SDouble(".0464348218674976747202318809261"),
         SDouble(".0509445739237286919327076700504"),
	 SDouble(".0551951053482859947448323724198"),
         SDouble(".0591114008806395723749672206486"),
	 SDouble(".0626532375547811680258701221743"),
         SDouble(".0658345971336184221115635569694"),
	 SDouble(".0686486729285216193456234118854"),
         SDouble(".0710544235534440683057903617232"),
	 SDouble(".0730306903327866674951894176589"),
         SDouble(".0745828754004991889865814183625"),
	 SDouble(".0757044976845566746595427753766"),
         SDouble(".0763778676720807367055028350381"),
	 SDouble(".0766007119179996564450499015301") };

static Double wg_41[10] =
       { SDouble(".0176140071391521183118619623519"),
         SDouble(".0406014298003869413310399522749"),
	 SDouble(".062672048334109063569506535187"),
         SDouble(".0832767415767047487247581432221"),
	 SDouble(".10193011981724043503675013548"),
         SDouble(".118194531961518417312377377711"),
	 SDouble(".131688638449176626898494499748"),
         SDouble(".142096109318382051329298325067"),
	 SDouble(".149172986472603746787828737002"),
         SDouble(".152753387130725850698084331955") };

static Double xgk_51[26] =
       { SDouble(".99926210499260983419345748654"),
         SDouble(".995556969790498097908784946894"),
	 SDouble(".988035794534077247637331014577"),
         SDouble(".97666392145951751149831538648"),
	 SDouble(".96161498642584251241813003366"),
         SDouble(".942974571228974339414011169659"),
	 SDouble(".920747115281701561746346084546"),
         SDouble(".894991997878275368851042006783"),
	 SDouble(".865847065293275595448996969588"),
         SDouble(".833442628760834001421021108694"),
	 SDouble(".797873797998500059410410904994"),
         SDouble(".759259263037357630577282865204"),
	 SDouble(".717766406813084388186654079773"),
         SDouble(".673566368473468364485120633248"),
	 SDouble(".626810099010317412788122681625"),
         SDouble(".577662930241222967723689841613"),
	 SDouble(".526325284334719182599623778158"),
         SDouble(".473002731445714960522182115009"),
	 SDouble(".417885382193037748851814394595"),
         SDouble(".361172305809387837735821730128"),
	 SDouble(".30308953893110783016747890998"),
         SDouble(".243866883720988432045190362798"),
	 SDouble(".18371893942104889201596988876"),
         SDouble(".122864692610710396387359818808"),
	 SDouble(".0615444830056850788865463923668"),
         SDouble("0.") };

static Double wgk_51[26] =
       { SDouble(".00198738389233031592650785188284"),
         SDouble(".00556193213535671375804023690107"),
         SDouble(".00947397338617415160720771052366"),
	 SDouble(".013236229195571674813656405847"),
         SDouble(".0168478177091282982315166675363"),
	 SDouble(".0204353711458828354565682922359"),
         SDouble(".0240099456069532162200924891649"),
	 SDouble(".0274753175878517378029484555178"),
         SDouble(".0307923001673874888911090202152"),
	 SDouble(".0340021302743293378367487952296"),
         SDouble(".0371162714834155435603306253676"),
	 SDouble(".0400838255040323820748392844671"),
         SDouble(".0428728450201700494768957924395"),
	 SDouble(".0455029130499217889098705847527"),
         SDouble(".0479825371388367139063922557569"),
	 SDouble(".0502776790807156719633252594334"),
         SDouble(".0523628858064074758643667121379"),
	 SDouble(".0542511298885454901445433704599"),
         SDouble(".0559508112204123173082406863828"),
	 SDouble(".0574371163615678328535826939395"),
         SDouble(".0586896800223942079619741758568"),
	 SDouble(".0597203403241740599790992919326"),
         SDouble(".0605394553760458629453602675176"),
	 SDouble(".0611285097170530483058590304163"),
         SDouble(".0614711898714253166615441319653"),
	 SDouble(".0615808180678329350787598242401")
         };

static Double wg_51[13] =
       { SDouble(".0113937985010262879479029641132"),
         SDouble(".0263549866150321372619018152953"),
	 SDouble(".0409391567013063126556234877117"),
         SDouble(".0549046959758351919259368915405"),
	 SDouble(".0680383338123569172071871856567"),
         SDouble(".0801407003350010180132349596691"),
	 SDouble(".0910282619829636498114972207029"),
         SDouble(".100535949067050644202206890393"),
	 SDouble(".10851962447426365311609395705"),
         SDouble(".11485825914571164833932554587"),
	 SDouble(".119455763535784772228178126513"),
         SDouble(".122242442990310041688959518946"),
	 SDouble(".123176053726715451203902873079") };

static Double xgk_61[31] =
       { SDouble(".999484410050490637571325895706"),
         SDouble(".996893484074649540271630050919"),
	 SDouble(".99163099687040459485862836611"),
         SDouble(".983668123279747209970032581606"),
	 SDouble(".973116322501126268374693868424"),
         SDouble(".960021864968307512216871025582"),
	 SDouble(".944374444748559979415831324037"),
         SDouble(".926200047429274325879324277081"),
	 SDouble(".905573307699907798546522558926"),
         SDouble(".88256053579205268154311646253"),
	 SDouble(".857205233546061098958658510659"),
         SDouble(".829565762382768397442898119733"),
	 SDouble(".799727835821839083013668942323"),
         SDouble(".767777432104826194917977340975"),
	 SDouble(".73379006245322680472617113137"),
         SDouble(".697850494793315796932292388027"),
	 SDouble(".660061064126626961370053668149"),
         SDouble(".620526182989242861140477556431"),
	 SDouble(".579345235826361691756024932173"),
         SDouble(".536624148142019899264169793311"),
	 SDouble(".492480467861778574993693061208"),
         SDouble(".447033769538089176780609900323"),
	 SDouble(".400401254830394392535476211543"),
         SDouble(".352704725530878113471037207089"),
	 SDouble(".304073202273625077372677107199"),
         SDouble(".254636926167889846439805129818"),
	 SDouble(".204525116682309891438957671002"),
         SDouble(".153869913608583546963794672743"),
	 SDouble(".102806937966737030147096751318"),
         SDouble(".0514718425553176958330252131667"),
	 SDouble("0.") };

static Double wgk_61[31] =
       { SDouble(".00138901369867700762455159122676"),
         SDouble(".00389046112709988405126720184452"),
         SDouble(".00663070391593129217331982636975"),
         SDouble(".00927327965951776342844114689202"),
         SDouble(".0118230152534963417422328988533"),
	 SDouble(".0143697295070458048124514324436"),
         SDouble(".0169208891890532726275722894203"),
	 SDouble(".0194141411939423811734089510501"),
         SDouble(".0218280358216091922971674857383"),
	 SDouble(".0241911620780806013656863707252"),
         SDouble(".0265099548823331016106017093351"),
	 SDouble(".0287540487650412928439787853543"),
         SDouble(".0309072575623877624728842529431"),
	 SDouble(".0329814470574837260318141910169"),
         SDouble(".0349793380280600241374996707315"),
	 SDouble(".0368823646518212292239110656171"),
         SDouble(".0386789456247275929503486515323"),
	 SDouble(".0403745389515359591119952797525"),
         SDouble(".041969810215164246147147541286"),
	 SDouble(".0434525397013560693168317281171"),
         SDouble(".0448148001331626631923555516167"),
	 SDouble(".0460592382710069881162717355594"),
         SDouble(".0471855465692991539452614781811"),
	 SDouble(".0481858617570871291407794922983"),
         SDouble(".0490554345550297788875281653672"),
	 SDouble(".0497956834270742063578115693799"),
         SDouble(".0504059214027823468408930856536"),
	 SDouble(".0508817958987496064922974730498"),
         SDouble(".0512215478492587721706562826049"),
	 SDouble(".0514261285374590259338628792158"),
         SDouble(".0514947294294515675583404336471") };

static Double wg_61[15] =
       { SDouble(".00796819249616660561546588347467"),
         SDouble(".0184664683110909591423021319121"),
	 SDouble(".0287847078833233693497191796113"),
         SDouble(".0387991925696270495968019364464"),
	 SDouble(".0484026728305940529029381404228"),
         SDouble(".0574931562176190664817216894021"),
	 SDouble(".065974229882180495128128515116"),
         SDouble(".0737559747377052062682438500222"),
	 SDouble(".0807558952294202153546949384605"),
         SDouble(".0868997872010829798023875307151"),
	 SDouble(".0921225222377861287176327070876"),
         SDouble(".0963687371746442596394686263518"),
	 SDouble(".0995934205867952670627802821036"),
         SDouble(".101762389748405504596428952169"),
	 SDouble(".102852652893558840341285636705") };

void QAGCLASS::init_qk15()
{
  xgk=xgk_15;
  wgk=wgk_15;
  wg=wg_15;
  even_gauss=FALSE;
  gauss_center=3;
  kron_center=7;
  loop1l=3;
  loop2l=4;
  loop3l=7;
}

void QAGCLASS::init_qk21()
{
  xgk=xgk_21;
  wgk=wgk_21;
  wg=wg_21;
  even_gauss=TRUE;
  gauss_center=0;
  kron_center=10;
  loop1l=5;
  loop2l=5;
  loop3l=10;
}

void QAGCLASS::init_qk31()
{
  xgk=xgk_31;
  wgk=wgk_31;
  wg=wg_31;
  even_gauss=FALSE;
  gauss_center=7;
  kron_center=15;
  loop1l=7;
  loop2l=8;
  loop3l=15;
}

void QAGCLASS::init_qk41()
{
  xgk=xgk_41;
  wgk=wgk_41;
  wg=wg_41;
  even_gauss=TRUE;
  gauss_center=0;
  kron_center=20;
  loop1l=10;
  loop2l=10;
  loop3l=20;
}

void QAGCLASS::init_qk51()
{
  xgk=xgk_51;
  wgk=wgk_51;
  wg=wg_51;
  even_gauss=FALSE;
  gauss_center=12;
  kron_center=25;
  loop1l=12;
  loop2l=13;
  loop3l=25;
}

void QAGCLASS::init_qk61()
{
  xgk=xgk_61;
  wgk=wgk_61;
  wg=wg_61;
  even_gauss=TRUE;
  gauss_center=0;
  kron_center=30;
  loop1l=15;
  loop2l=15;
  loop3l=30;
}



/* ---------------------------------------------------------------- */
/*                qk: general routine for gauss-kronrod integration */
/* ---------------------------------------------------------------- */

RESULT_TYPE QAGCLASS::qk(RESULT_TYPE (*f)(Double),Double a,Double b,
                    Double *abserr,Double *resabs,Double *resasc)
{
    RESULT_TYPE result,resg,resk,fc,fsum,fval1,fval2,reskh;

    Double absc, x, hlgth, centr, dhlgth;
    int j,jtw,jtwm1;

    centr = (a + b) / 2;
    hlgth = (b - a) / 2;
    dhlgth = fabs(hlgth);
/*                           COMPUTE THE KRONROD                */
/*                           APPROXIMATION TO THE INTEGRAL, AND */
/*                           ESTIMATE THE ABSOLUTE ERROR.       */
    fc = (*f)(centr);
    if (even_gauss) resg=RESULT_TYPE(0); else resg=fc*wg[gauss_center];
    resk = fc * wgk[kron_center];
    *resabs = abs(resk);
    for (j = 0; j < loop1l; j++) {
     jtw = j*2+1;
     absc = hlgth*xgk[jtw];
     x = centr-absc;
     fval1 = (*f)(x);
     x = centr + absc;
     fval2 = (*f)(x);
     fv1[jtw] = fval1;
     fv2[jtw] = fval2;
     fsum = fval1 + fval2;
     resg += wg[j] * fsum;
     resk += wgk[jtw] * fsum;
     *resabs += wgk[jtw] * (abs(fval1) + abs(fval2));
    };
    for (j = 0; j < loop2l; j++) {
     jtwm1 = j*2;
     absc = hlgth * xgk[jtwm1];
     x = centr - absc;
     fval1 = (*f)(x);
     x = centr + absc;
     fval2 = (*f)(x);
     fv1[jtwm1] = fval1;
     fv2[jtwm1] = fval2;
     fsum = fval1 + fval2;
     resk += wgk[jtwm1] * fsum;
     *resabs += wgk[jtwm1] * (abs(fval1) + abs(fval2));
    };
    reskh = resk /2 ;
    *resasc = wgk[kron_center] * abs(fc - reskh);
    for (j = 0; j < loop3l; j++)
     *resasc+=wgk[j] * (abs(fv1[j]-reskh)+abs(fv2[j]-reskh));
    result = resk * hlgth;
    *resabs *= dhlgth;
    *resasc *= dhlgth;
    *abserr = abs((resk - resg) * hlgth);
    if (*resasc != Double(0) && *abserr != Double(0)) 
     *abserr = *resasc * min(1.0 , pow(*abserr * 200. / *resasc, ERR_EXP));
    if (*resabs > UFLOW / EPMACH50 )
     *abserr = max(EPMACH50 * (*resabs),*abserr);
    return(result);
} /* qk */

RESULT_TYPE QAGCLASS::qk15(RESULT_TYPE (*f)(Double),Double a,Double b,
                      Double *abserr,Double *resabs,Double *resasc)
{
  init_qk15();
  return(qk(f,a,b,abserr,resabs,resasc));
}

RESULT_TYPE QAGCLASS::qk21(RESULT_TYPE (*f)(Double),Double a,Double b,
                      Double *abserr,Double *resabs,Double *resasc)
{
  init_qk21();
  return(qk(f,a,b,abserr,resabs,resasc));
}

RESULT_TYPE QAGCLASS::qk31(RESULT_TYPE (*f)(Double),Double a,Double b,
                      Double *abserr,Double *resabs,Double *resasc)
{
  init_qk31();
  return(qk(f,a,b,abserr,resabs,resasc));
}

RESULT_TYPE QAGCLASS::qk41(RESULT_TYPE (*f)(Double),Double a,Double b,
                      Double *abserr,Double *resabs,Double *resasc)
{
  init_qk41();
  return(qk(f,a,b,abserr,resabs,resasc));
}

RESULT_TYPE QAGCLASS::qk51(RESULT_TYPE (*f)(Double),Double a,Double b,
                      Double *abserr,Double *resabs,Double *resasc)
{
  init_qk51();
  return(qk(f,a,b,abserr,resabs,resasc));
}

RESULT_TYPE QAGCLASS::qk61(RESULT_TYPE (*f)(Double),Double a,Double b,
                      Double *abserr,Double *resabs,Double *resasc)
{
  init_qk61();
  return(qk(f,a,b,abserr,resabs,resasc));
}



void QAGCLASS::qsort(int limit,int last,int *maxerr,
                Double *ermax,Double *elist,int *iord,int *nrmax)
{
    Double errmax, errmin;
    int i, ibeg, ido, isucc, j, jbnd, jupbn, k;

    if (last <=2 ) {
      iord[0] = 1;
      iord[1] = 2;
      *maxerr = iord[*nrmax-1];
      *ermax = elist[*maxerr-1];
      return; };
/*                                  THIS PART OF THE ROUTINE IS ONLY     */
/*                                  EXECUTED IF, DUE TO A DIFFICULT      */
/*                                  INTEGRAND, SUBDIVISION INCREASED THE */
/*                                  ERROR ESTIMATE.  IN THE NORMAL CASE  */
/*                                  THE INSERT PROCEDURE SHOULD START    */
/*                                  AFTER THE NRMAX-TH LARGEST ERROR     */
/*                                  ESTIMATE                             */
    errmax = elist[*maxerr-1];
    if (*nrmax != 1) {
      ido = *nrmax - 1;
      for (i = 1; i <= ido; ++i) {
        isucc = iord[*nrmax - 2];
/*                                  JUMP OUT OF LOOP */
        if (errmax <= elist[isucc-1]) break;
        iord[*nrmax-1] = isucc;
        --(*nrmax); } }
/*                                  COMPUTE THE NUMBER OF ELEMENTS IN THE */
/*                                  LIST TO BE MAINTAINED IN DESCENDING   */
/*                                  ORDER.  THIS NUMBER DEPENDS ON THE    */
/*                                  NUMBER OF SUBDIVISIONS STILL ALLOWED  */
    jupbn = last;
    if (last > limit / 2 + 2) jupbn = limit + 3 - last;
    errmin = elist[last-1];
/*                                  INSERT ERRMAX BY TRAVERSING THE LIST */
/*                                  TOP-DOWN, STARTING COMPARISON FROM   */
/*                                  THE ELEMENT ELIST(IORD(NRMAX+1))     */
    jbnd = jupbn - 1;
    ibeg = *nrmax + 1;
    if (ibeg <= jbnd) {
      for (i = ibeg; i <= jbnd; ++i) {
        isucc = iord[i-1];
        if (errmax >= elist[isucc-1]) {
          iord[i - 2] = *maxerr;
          k = jbnd;
          for (j = i; j <= jbnd; ++j) {
            isucc = iord[k-1];
/*                                  JUMP OUT OF LOOP */
            if (errmin < elist[isucc-1]) { i=k+1; break; };
            iord[k] = isucc;
            --k; }
          iord[i-1] = last;
/*                                  SET MAXERR AND ERMAX. */
          *maxerr = iord[*nrmax-1];
          *ermax = elist[*maxerr-1];
          return; }
        iord[i - 2] = isucc; } };
    iord[jbnd-1] = *maxerr;
    iord[jupbn-1] = last;
    *maxerr = iord[*nrmax-1];
    *ermax = elist[*maxerr-1];
    return;
/*                                  INSERT ERRMIN BY TRAVERSING THE LIST */
/*                                  BOTTOM-UP                            */
} /* qsort */




/***************************************************************************/

/* 1.     QAGE */
/*        COMPUTATION OF A DEFINITE INTEGRAL */
/*           STANDARD FORTRAN SUBROUTINE */

/* 2.     PURPOSE */
/*           THE ROUTINE CALCULATES AN APPROXIMATION  RESULT  TO A GIVEN */
/*           DEFINITE INTEGRAL   I = INTEGRAL OF  F  OVER (A,B), */
/*           HOPEFULLY SATISFYING FOLLOWING CLAIM FOR ACCURACY */
/*           ABS(I-RESULT).LE.MAX(EPSABS,EPSREL*ABS(I)). */

/* 3.     CALLING SEQUENCE */
/*           CALL QAGE(F,A,B,EPSABS,EPSREL,KEY,LIMIT,RESULT,ABSERR, */
/*                      NEVAL,IER,ALIST,BLIST,RLIST,ELIST,IORD,LAST) */

/*        PARAMETERS */
/*         ON ENTRY */
/*            F      - REAL */
/*                     FUNCTION SUBPROGRAM DEFINING THE INTEGRAND */
/*                     FUNCTION F(X). THE ACTUAL NAME FOR F NEEDS TO BE */
/*                     DECLARED E X T E R N A L IN THE DRIVER PROGRAM. */

/*            A      - REAL */
/*                     LOWER LIMIT OF INTEGRATION */

/*            B      - REAL */
/*                     UPPER LIMIT OF INTEGRATION */

/*            EPSABS - REAL */
/*                     ABSOLUTE ACCURACY REQUESTED */
/*            EPSREL - REAL */
/*                     RELATIVE ACCURACY REQUESTED */
/*                     IF  EPSABS.LT.0 AND EPSREL.LT.0, */
/*                     THE ROUTINE WILL END WITH IER = 6. */

/*            KEY    - INTEGER */
/*                     KEY FOR CHOICE OF LOCAL INTEGRATION RULE */
/*                     A GAUSS-KRONROD PAIR IS USED WITH */
/*                          7 - 15 POINTS IF KEY.LT.2, */
/*                         10 - 21 POINTS IF KEY = 2, */
/*                         15 - 31 POINTS IF KEY = 3, */
/*                         20 - 41 POINTS IF KEY = 4, */
/*                         25 - 51 POINTS IF KEY = 5, */
/*                         30 - 61 POINTS IF KEY.GT.5. */

/*            LIMIT  - INTEGER */
/*                     GIVES AN UPPERBOUND ON THE NUMBER OF SUBINTERVALS */
/*                     IN THE PARTITION OF (A,B), LIMIT.GE.1. */

/*         ON RETURN */
/*            RESULT - REAL */
/*                     APPROXIMATION TO THE INTEGRAL */

/*            ABSERR - REAL */
/*                     ESTIMATE OF THE MODULUS OF THE ABSOLUTE ERROR, */
/*                     WHICH SHOULD EQUAL OR EXCEED ABS(I-RESULT) */

/*            NEVAL  - INTEGER */
/*                     NUMBER OF INTEGRAND EVALUATIONS */

/*            IER    - INTEGER */
/*                     IER = 0 NORMAL AND RELIABLE TERMINATION OF THE */
/*                             ROUTINE. IT IS ASSUMED THAT THE REQUESTED */
/*                             ACCURACY HAS BEEN ACHIEVED. */
/*                     IER.GT.0 ABNORMAL TERMINATION OF THE ROUTINE */
/*                             THE ESTIMATES FOR RESULT AND ERROR ARE */
/*                             LESS RELIABLE. IT IS ASSUMED THAT THE */
/*                             REQUESTED ACCURACY HAS NOT BEEN ACHIEVED. */
/*                     IER = 1 MAXIMUM NUMBER OF SUBDIVISIONS ALLOWED */
/*                             HAS BEEN ACHIEVED. ONE CAN ALLOW MORE */
/*                             SUBDIVISIONS BY INCREASING THE VALUE */
/*                             OF LIMIT. */
/*                             HOWEVER, IF THIS YIELDS NO IMPROVEMENT IT */
/*                             IS RATHER ADVISED TO ANALYZE THE INTEGRAND */
/*                             IN ORDER TO DETERMINE THE INTEGRATION */
/*                             DIFFICULTIES. IF THE POSITION OF A LOCAL */
/*                             DIFFICULTY CAN BE DETERMINED(E.G. */
/*                             SINGULARITY, DISCONTINUITY WITHIN THE */
/*                             INTERVAL) ONE WILL PROBABLY GAIN FROM */
/*                             SPLITTING UP THE INTERVAL AT THIS POINT */
/*                             AND CALLING THE INTEGRATOR ON THE */
/*                             SUBRANGES. IF POSSIBLE, AN APPROPRIATE */
/*                             SPECIAL-PURPOSE INTEGRATOR SHOULD BE USED */
/*                             WHICH IS DESIGNED FOR HANDLING THE TYPE OF */
/*                             DIFFICULTY INVOLVED. */
/*                         = 2 THE OCCURRENCE OF ROUNDOFF ERROR IS */
/*                             DETECTED, WHICH PREVENTS THE REQUESTED */
/*                             TOLERANCE FROM BEING ACHIEVED. */
/*                         = 3 EXTREMELY BAD INTEGRAND BEHAVIOUR OCCURS */
/*                             AT SOME POINTS OF THE INTEGRATION */
/*                             INTERVAL. */
/*                         = 6 THE INPUT IS INVALID, BECAUSE */
/*                             EPSABS.LT.0 AND EPSREL.LT.0, */
/*                             RESULT, ABSERR, NEVAL, LAST, RLIST(1), */
/*                             ELIST(1) AND IORD(1) ARE SET TO ZERO. */
/*                             ALIST(1) AND BLIST(1) ARE SET TO A AND B */
/*                             RESPECTIVELY. */

/*            ALIST   - REAL */
/*                      VECTOR OF DIMENSION AT LEAST LIMIT, THE FIRST */
/*                       LAST  ELEMENTS OF WHICH ARE THE LEFT END POINTS */
/*                      OF THE SUBINTERVALS IN THE PARTITION OF THE */
/*                      GIVEN INTEGRATION RANGE (A,B) */

/*            BLIST   - REAL */
/*                      VECTOR OF DIMENSION AT LEAST LIMIT, THE FIRST */
/*                       LAST  ELEMENTS OF WHICH ARE THE RIGHT END POINTS */
/*                      OF THE SUBINTERVALS IN THE PARTITION OF THE */
/*                      GIVEN INTEGRATION RANGE (A,B) */

/*            RLIST   - REAL */
/*                      VECTOR OF DIMENSION AT LEAST LIMIT, THE FIRST */
/*                       LAST  ELEMENTS OF WHICH ARE THE INTEGRAL */
/*                      APPROXIMATIONS ON THE SUBINTERVALS */

/*            ELIST   - REAL */
/*                      VECTOR OF DIMENSION AT LEAST LIMIT, THE FIRST */
/*                       LAST  ELEMENTS OF WHICH ARE THE MODULI OF THE */
/*                      ABSOLUTE ERROR ESTIMATES ON THE SUBINTERVALS */

/*            IORD    - INTEGER */
/*                      VECTOR OF DIMENSION AT LEAST LIMIT, THE FIRST K */
/*                      ELEMENTS OF WHICH ARE POINTERS TO THE ERROR */
/*                      ESTIMATES OVER THE SUBINTERVALS, SUCH THAT */
/*                      ELIST(IORD(1)), ..., ELIST(IORD(K)) FORM A */
/*                      DECREASING SEQUENCE, WITH */
/*                      K = LAST IF LAST.LE.(LIMIT/2+2), AND */
/*                      K = LIMIT+1-LAST OTHERWISE */

/*            LAST    - INTEGER */
/*                      NUMBER OF SUBINTERVALS ACTUALLY PRODUCED IN THE */
/*                      SUBDIVISION PROCESS */

/* 4.     SUBROUTINES OR FUNCTIONS NEEDED */
/*              - QK15, QK21, QK31, QK41, QK51, QK61 */
/*              - QSORT */
/*              - F (USER-PROVIDED FUNCTION) */
/*              - QMACO */
/*              - FORTRAN ABS, AMAX1, AMIN1 */

/**************************************************************************/




/*            LIST OF MAJOR VARIABLES */
/*            ----------------------- */

/*           ALIST     - LIST OF LEFT END POINTS OF ALL SUBINTERVALS */
/*                       CONSIDERED UP TO NOW */
/*           BLIST     - LIST OF RIGHT END POINTS OF ALL SUBINTERVALS */
/*                       CONSIDERED UP TO NOW */
/*           RLIST(I)  - APPROXIMATION TO THE INTEGRAL OVER */
/*                      (ALIST(I),BLIST(I)) */
/*           ELIST(I)  - ERROR ESTIMATE APPLYING TO RLIST(I) */
/*           MAXERR    - POINTER TO THE INTERVAL WITH LARGEST */
/*                       ERROR ESTIMATE */
/*           ERRMAX    - ELIST(MAXERR) */
/*           AREA      - SUM OF THE INTEGRALS OVER THE SUBINTERVALS */
/*           ERRSUM    - SUM OF THE ERRORS OVER THE SUBINTERVALS */
/*           ERRBND    - REQUESTED ACCURACY MAX(EPSABS,EPSREL* */
/*                       ABS(RESULT)) */
/*           *****1    - VARIABLE FOR THE LEFT SUBINTERVAL */
/*           *****2    - VARIABLE FOR THE RIGHT SUBINTERVAL */
/*           LAST      - INDEX FOR SUBDIVISION */


/*           MACHINE DEPENDENT CONSTANTS */
/*           --------------------------- */

/*           EPMACH  IS THE LARGEST RELATIVE SPACING. */
/*           UFLOW  IS THE SMALLEST POSITIVE MAGNITUDE. */
/*           OFLOW  IS THE LARGEST MAGNITUDE. */

/* ----------------------------------------------------------------------- */

RESULT_TYPE QAGCLASS::qage(RESULT_TYPE (*f)(Double),Double a,Double b,Double epsabs,
                      Double epsrel,int key,int limit,
                      Double *abserr,int *neval,int *ier,Double *alist,
                      Double *blist,RESULT_TYPE *rlist,Double *elist,
                      int *iord,int *last)
{
    RESULT_TYPE result,area,area1,area2,area12;
    int keyf;
    Double c;
    int k;
    Double erro12, a1, a2, b1;
    int nrmax;
    Double b2, defab1, defab2;
    int iroff1, iroff2;
    Double error1, error2, defabs, errbnd, resabs;
    int maxerr;
    Double errmax, errsum;


/*                                  TEST ON VALIDITY OF PARAMETERS */

    *ier = 0;
    *neval = 0;
    *last = 0;
    result = RESULT_TYPE(0);
    *abserr = 0.;
    alist[0] = a;
    blist[0] = b;
    rlist[0] = RESULT_TYPE(0);
    elist[0] = Double(0);
    iord[0] = 0;
    if (epsabs < 0. && epsrel < 0.) {
      *ier = 6; return(result); };

/*                                  FIRST APPROXIMATION TO THE INTEGRAL */
    keyf = key;
    if (key <= 0) {
     keyf = 1;
    }
    if (key >= 7) {
     keyf = 6;
    }
    c = (Double) keyf;
    *neval = 0;
    if (keyf == 1) init_qk15();
    if (keyf == 2) init_qk21();
    if (keyf == 3) init_qk31();
    if (keyf == 4) init_qk41();
    if (keyf == 5) init_qk51();
    if (keyf == 6) init_qk61();
    result = qk(f, a, b, abserr, &defabs, &resabs);
    *last = 1;
    rlist[0] = result;
    elist[0] = *abserr;
    iord[0] = 1;

/*                                  TEST ON ACCURACY. */

/* Computing MAX */
    errbnd = max(epsabs, epsrel * abs(result));
    if (*abserr <= EPMACH50 * defabs && *abserr > errbnd) *ier = 2;
    if (limit == 1) *ier = 1;
    if (*ier != 0 || *abserr<=errbnd && *abserr != resabs || *abserr == 0.) {
      if (keyf != 1) {
        *neval = (keyf * 10 + 1) * ((*neval << 1) + 1);
      }
      if (keyf == 1) {
      *neval = *neval * 30 + 15;
      }
      return(result); }

/*                                  INITIALIZATION */

    errmax = *abserr;
    maxerr = 1;
    area = result;
    errsum = *abserr;
    nrmax = 1;
    iroff1 = 0;
    iroff2 = 0;

/*                                  MAIN LOOP */

    for (*last = 2; (*last <= limit); ++(*last)) {

/*                                  BISECT THE SUBINTERVAL WITH THE */
/*                                  LARGEST ERROR ESTIMATE */

     a1 = alist[maxerr-1];
     b1 = (alist[maxerr-1] + blist[maxerr-1]) / 2;
     a2 = b1;
     b2 = blist[maxerr-1];
     area1=qk(f, a1, b1, &error1, &resabs, &defab1);
     area2=qk(f, a2, b2, &error2, &resabs, &defab2);

/*                                  IMPROVE PREVIOUS APPROXIMATIONS TO */
/*                                  INTEGRAL AND ERROR;  TEST FOR */
/*                                  ACCURACY */

     ++(*neval);
     area12 = area1 + area2;
     erro12 = error1 + error2;
     errsum = errsum + erro12 - errmax;
     area = area + area12 - rlist[maxerr-1];
     if (defab1 != error1 && defab2 != error2) {
       if ((abs(rlist[maxerr-1] - area12)) <= abs(area12) * 1e-5 
            && erro12 >= errmax * .99) ++iroff1;
       if (*last > 10 && erro12 > errmax) ++iroff2; };
     rlist[maxerr-1] = area1;
     rlist[*last-1] = area2;
/* Computing MAX */
     errbnd = max(epsabs, epsrel * abs(area));
     if (errsum > errbnd) {

/*                                  TEST FOR ROUNDOFF ERROR AND   */
/*                                  EVENTUALLY SET ERROR FLAG     */

#ifdef QUADPRECISION
     if (iroff1 >= 12 || iroff2 >= 40) *ier = 2 ;    // experimental !!
#else
     if (iroff1 >= 6 || iroff2 >= 20) *ier =2;
#endif

/*                                  SET ERROR FLAG IN THE CASE THAT THE */
/*                                  NUMBER OF SUBINTERVALS EQUALS LIMIT */

     if (*last == limit) *ier = 1;

/*                                  SET ERROR FLAG IN THE CASE OF BAD     */
/*                                  INTEGRAND BEHAVIOR AT A POINT OF THE  */
/*                                  INTEGRATION RANGE                     */

/* Computing MAX */
//
//   This part of the code produced unnecessary exits. How to fix it?
//
//     if (max(abs(a1),abs(b2)) <= (EPMACH*100.+1.) * (abs(a2)+UFLOW*1e3)) {
//         *ier = 3;
//     }


/*                                  APPEND THE NEWLY-CREATED INTERVALS */
/*                                  TO THE LIST                        */

     }
     if (error2 <= error1) {
       alist[*last-1] = a2;
       blist[maxerr-1] = b1;
       blist[*last-1] = b2;
       elist[maxerr-1] = error1;
       elist[*last-1] = error2; }
     else { 
       alist[maxerr-1] = a2;
       alist[*last-1] = a1;
       blist[*last-1] = b1;
       rlist[maxerr-1] = area2;
       rlist[*last-1] = area1;
       elist[maxerr-1] = error2;
       elist[*last-1] = error1; }

/*                                  CALL SUBROUTINE QSORT TO MAINTAIN   */
/*                                  THE DESCENDING ORDERING IN THE LIST */
/*                                  OF ERROR ESTIMATES AND SELECT THE   */
/*                                  SUBINTERVAL WITH THE LARGEST ERROR  */
/*                                  ESTIMATE (TO BE BISECTED NEXT)      */

     qsort(limit, *last, &maxerr, &errmax, &elist[0], &iord[0], &nrmax);
/*                                  JUMP OUT OF LOOP */
     if (*ier != 0 || errsum <= errbnd) break;
   }

/*                                  COMPUTE FINAL RESULT. */

    result = RESULT_TYPE(0);
    for (k = 1; k <= *last; ++k) {
     result += rlist[k-1];
    }
    *abserr = errsum;
    if (keyf != 1) {
     *neval = (keyf * 10 + 1) * ((*neval << 1) + 1);
    }
    if (keyf == 1) {
     *neval = *neval * 30 + 15;
    }
    return(result);
} /* qage */


RESULT_TYPE QAGCLASS::qagmain(RESULT_TYPE (*f) (Double),Double a,Double b,
                              Double epsabs,
                              Double epsrel,int key,Double *abserr,
                              int *neval,int *ier,int limit,int noisy)
{

    int last;
    RESULT_TYPE result;

    if (noisy==2) Outstream << "Integration routine QAG called.\n";
    result=qage(f, a, b, epsabs, epsrel, key, limit, abserr, neval, ier, 
                alist, blist, rlist, elist, iord, &last);

    if (*ier > 0) {
      if ((*ier == 1)&&(noisy>=1)&&(evaluation_status != FIX_SUBDIVISIONS))
        Outstream << "Integration problem: \n"
             << "MAXIMUM NUMBER OF SUBDIVISIONS ALLOWED HAS BEEN ACHIEVED.\n";
      else {
      if ((*ier == 2)&&(noisy>=1))
        Outstream << "Integration problem: \n"
             << "OCCURRENCE OF ROUNDOFF ERROR.\n";
      else {
      if ((*ier == 3)&&(noisy>=1))
        Outstream << "Integration problem: \n"
             << "EXTREMELY BAD INTEGRAND BEHAVIOUR AT SOME POINTS"
             << " OF THE INTEGRATION INTERVAL.\n"
             << "subdivisions = " << *neval << "\n";
      else {
      if (*ier == 6)                          // &&(noisy==1)
        Outstream << "Integration problem: \n"
             << "INVALID INPUT.\n";
      else if (noisy==1) cout << "Integration problem\n";
      };};};};
    return(result);
} /* qag */



void QAGCLASS::initialize()
{
  error_handling=ERROR_REPORT;
  evaluation_status=FIX_SUBDIVISIONS;
  max_subdivisions=50;
  max_absolute_error=0.0;
  max_relative_error=0.000000001;
  standard_key=2;
  reported_absolute_error=0;
  reported_relative_error=0;
  reported_f_evaluations=0;
  reported_error_status=0;
}

void QAGCLASS::create()
{
  initialize();
  fv1 = new RESULT_TYPE[30];
  fv2 = new RESULT_TYPE[30];
  iord = new int[QMAXINT];
  alist = new Double[QMAXINT];
  blist = new Double[QMAXINT];
  elist = new Double[QMAXINT];
  rlist = new RESULT_TYPE[QMAXINT];
}

QAGCLASS::QAGCLASS(void) { create(); }

QAGCLASS::QAGCLASS(int error_type,Double error)
{
  create();
  if (error<0) {"Wrong initialisation of qag\n"; return;};
  max_subdivisions=QMAXINT;
  if (error_type==FIX_RELATIVE_ERROR) max_relative_error=error;
  else {
    if (error_type==FIX_ABSOLUTE_ERROR) max_absolute_error=error;
    else Outstream << "Wrong initialisation of qag: error_type\n"; };
}

QAGCLASS::QAGCLASS(Double abs_error,Double rel_error,int max_subdiv)
{
  create();
  if ((abs_error<0)||(rel_error<0)||(max_subdiv<0)||(max_subdiv>=QMAXINT)) {
    Outstream <<  "Wrong initialisation of qag\n"; }
  else {
    max_absolute_error=abs_error;
    max_relative_error=rel_error;
    max_subdivisions=max_subdiv; };
}

void QAGCLASS::set(Double abs_error,Double rel_error,int max_subdiv)
{
  initialize();
  if ((abs_error<0)||(rel_error<0)||(max_subdiv<0)||(max_subdiv>=QMAXINT)) {
    Outstream <<  "Wrong initialisation of qag\n"; }
  else {
    max_absolute_error=abs_error;
    max_relative_error=rel_error;
    max_subdivisions=max_subdiv; };
}

QAGCLASS::~QAGCLASS(void)
{
  delete fv1;
  delete fv2;
  delete iord;
  delete alist;
  delete blist;
  delete elist;
  delete rlist;
}


void QAGCLASS::set_error_handling(int i) { error_handling=i;}


RESULT_TYPE QAGCLASS::integrate(RESULT_TYPE (*f) (Double),Double a,Double b)
{
  RESULT_TYPE r;
  r=qagmain(f,a,b,max_absolute_error,max_relative_error,standard_key,
                 &reported_absolute_error,&reported_f_evaluations,
                 &reported_error_status,max_subdivisions,error_handling);
  reported_relative_error=abs(reported_absolute_error/r);
  return(r);
}

void QAGCLASS::reported_errors(Double *relative,Double *absolute,
                               int *f_evaluations)
{
  *absolute=reported_absolute_error;
  *relative=reported_relative_error;
  *f_evaluations=reported_f_evaluations;
}

